body{margin:0;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;overflow-x:auto}code{font-family:source-code-pro,Menlo,Monaco,Consolas,"Courier New",monospace}.inner-container{width:110%;max-width:1920px;padding-top:82px;padding-left:250px;padding-bottom:10px}.vertical-tab-override,.vertical-tab-override .ui.grid{min-height:100vh}.vertical-tab-override .ui.vertical.menu.fluid{height:100%}.vertical-tab-override .ui.vertical.tabular.menu .item{display:flex;flex-direction:column;align-items:flex-start;justify-content:space-between;flex-wrap:wrap}.vertical-tab-override .ui.vertical.tabular.menu .item>.label{margin:.5rem 0 0}.vertical-tab-override .ui.attached.segment{border:none;padding:0}.button-link{background:none!important;border:none;padding:0!important;color:#4183c4;cursor:pointer;font-weight:700}#new_policy_wizard .ui.category.search .results{width:auto}.ui.basic.segment{width:90%}.ui.accordion,.ui.accordion .accordion{margin-bottom:1em}.ui.steps .active.step:after,.ui.steps .step:after{display:none}.ui.steps{border:none}.ui.steps .step{background:#fff;border-right:0}.ui.steps .step.active,.ui.steps:hover .step.active:hover{background:#fff}.step1:hover,.step2:hover,.step3:hover{background:#fff!important}.ui.steps .step .description{color:#686868}.ui.modal>.header{border-bottom:0}.ui.modal>.message{margin-top:0}.advancedOption>.menu.transition,.advancedOption>i{display:none!important}.helpContainer{position:relative}.help{display:inline-block;margin-bottom:0;position:absolute;bottom:0}.ui.items{border:1px solid rgba(34,36,38,.15);border-radius:.28571429rem}.ui.divided.items>.item:first-child .button{border-radius:0 .28571429rem 0 0}.ui.divided.items>.item:not(first-child) .button{border-radius:0}.ui.items>.item>.content>.header:not(.ui){width:100%}.ui.divided.items>.item:not(first-child){padding-top:0}.ui.items>.item .extra>.cogs.icon.actions{margin-bottom:0!important;font-size:1.25em;color:rgba(0,0,0,.85)}@media (max-width:1067px){.ui.fluid.search .results{width:150%}}.ui.steps .step>.icon{font-size:4.5em}.active.step1>.icon{font-family:Step;content:url(/static/media/stepper1Active.9866ff3a.svg)}.complete.step1>.icon{font-size:4em;font-family:Step;content:url(/static/media/stepper1Complete.6222088c.svg)}.step2>.icon{font-size:4em;font-family:Step;content:url(/static/media/stepper2Pending.3aac534f.svg)}.active.step2>.icon{font-family:Step;content:url(/static/media/stepper2Active.b5befeb5.svg)}.complete.step2>.icon{font-size:4em;font-family:Step;content:url(/static/media/stepper2Complete.47bfe6ee.svg)}.step3>.icon{font-size:4em;font-family:Step;content:url(/static/media/stepper3Pending.b72269cd.svg)}.active.step3>.icon{font-size:4em;font-family:Step;content:url(/static/media/stepper3Active.45f15c88.svg)}.users.icon,.users.large.icon{color:#286f85}.ui.blue.label,.ui.blue.labels .label{background-color:#286f85!important;border-color:#286f85!important}.infoError{background:rgba(33,133,208,.4)}.warningError{background:rgba(203,174,172,.4)}.criticalError{background:rgba(255,101,80,.4)}.warningIcon{display:block;background-image:url(/static/media/warning.5b4ee25d.png);background-size:contain;background-repeat:no-repeat;padding:2px;background-origin:content-box;background-position:bottom 0 right 4px}.sq-title-required{color:#db2828}.sq-hidden-custom{display:none}
/*# sourceMappingURL=main.a2aebff7.chunk.css.map */